﻿DECLARE @ArchivationInterval AS INT = 1;
DECLARE @ArchivationIntervalType AS BIT = 1; -- 0 = Day, 1 = Month
DECLARE @ArchivationIntervalMax AS INT = 12; -- 12 units of interval

DELETE FROM [nan].[Setting] WHERE Setting IN ('ArchivationInterval', 'ArchivationIntervalType', 'ArchivationIntervalMax');

IF NOT EXISTS (SELECT * FROM [nan].[Setting] WHERE Setting = 'PrintJobArchiveDate')
BEGIN
	EXEC [nan].[FillValuesForArchivator];
END

INSERT INTO [nan].[Setting] (Setting, Value)
VALUES	
	('ArchivationInterval', CAST(@ArchivationInterval AS NVARCHAR(50))),
	('ArchivationIntervalType', CAST(@ArchivationIntervalType AS NVARCHAR(50))),
	('ArchivationIntervalMax', CAST(@ArchivationIntervalMax AS NVARCHAR(50)));
	
DECLARE @Now AS DATETIME = GETUTCDATE();

DECLARE @ArchivationIntervalStart AS DATE = (
	SELECT 
		CASE @ArchivationIntervalType 
			WHEN 0 THEN DATEADD(DAY, -@ArchivationInterval * @ArchivationIntervalMax, @Now) 
			ELSE DATEADD(MONTH, -@ArchivationInterval * @ArchivationIntervalMax, DATEADD(DAY, -DAY(@Now) + 1, @Now)) 
		END);

DECLARE @PartitionDates AS TABLE (PartitionDate DATETIME);

WITH Dates (PartitionDate) AS 
(
	SELECT @ArchivationIntervalStart

	UNION ALL

	SELECT 
		CASE @ArchivationIntervalType 
			WHEN 0 THEN DATEADD(DAY, @ArchivationInterval, PartitionDate) 
			ELSE DATEADD(MONTH, @ArchivationInterval, PartitionDate) 
		END AS PartDate
	FROM Dates
	WHERE 
		CASE @ArchivationIntervalType 
			WHEN 0 THEN DATEADD(DAY, @ArchivationInterval, PartitionDate) 
			ELSE DATEADD(MONTH, @ArchivationInterval, PartitionDate) 
		END <= @Now
)
INSERT INTO @PartitionDates
SELECT * 
FROM Dates
OPTION (MAXRECURSION 1000);

DECLARE @Query NVARCHAR(4000) = (SELECT STUFF((SELECT ', ''' + CONVERT(NVARCHAR(50), PartitionDate, 112) + '''' FROM @PartitionDates FOR XML PATH('')), 1, 2, ''));
SET @Query = 'CREATE PARTITION FUNCTION ArchivatorPF (DATETIME) AS RANGE RIGHT FOR VALUES (' + @Query + ')';
EXEC sp_executesql @Query;

CREATE PARTITION SCHEME ArchivatorPS AS PARTITION ArchivatorPF ALL TO ([PRIMARY]);

IF (SELECT COUNT(*) FROM sys.indexes WHERE name = 'IXC_PrintJob_EventId_BeginPrintTime') > 0
BEGIN
	DROP INDEX [IXC_PrintJob_EventId_BeginPrintTime] ON [nan].[PrintJob];

	CREATE CLUSTERED INDEX [IXC_PrintJob_LastUsed] ON [nan].[PrintJob]
	(
		LastUsed
	) WITH (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, SORT_IN_TEMPDB = OFF, ONLINE = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
	ON ArchivatorPS(LastUsed);
END

IF (SELECT COUNT(*) FROM sys.indexes WHERE name = 'IXC_Event_ParentOid_Created') > 0
BEGIN
	DROP INDEX [IXC_Event_ParentOid_Created] ON [nan].[Event];

	CREATE CLUSTERED INDEX [IXC_Event_LastUsed] ON [nan].[Event]
	(
		LastUsed
	) WITH (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, SORT_IN_TEMPDB = OFF, ONLINE = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
	ON ArchivatorPS(LastUsed);
END

ALTER TABLE [nan].[Variable] DROP CONSTRAINT [FK_Variable_LabelJob];

ALTER TABLE [nan].[LabelJob] DROP CONSTRAINT [PK_LabelJob];

CREATE CLUSTERED INDEX [IXC_LabelJob_LastUsed] ON [nan].[LabelJob]
(
	LastUsed
) WITH (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, SORT_IN_TEMPDB = OFF, ONLINE = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
ON ArchivatorPS(LastUsed);

ALTER TABLE [nan].[LabelJob] ADD CONSTRAINT [PK_LabelJob] PRIMARY KEY NONCLUSTERED 
(
	[OID] ASC
)WITH (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, SORT_IN_TEMPDB = OFF, IGNORE_DUP_KEY = OFF, ONLINE = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
ON [PRIMARY];

ALTER TABLE [nan].[Variable] DROP CONSTRAINT [PK_Variable];

CREATE CLUSTERED INDEX [IXC_Variable_LastUsed] ON [nan].[Variable]
(
	LastUsed
) WITH (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, SORT_IN_TEMPDB = OFF, ONLINE = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
ON ArchivatorPS(LastUsed);

ALTER TABLE [nan].[Variable] ADD  CONSTRAINT [PK_Variable] PRIMARY KEY NONCLUSTERED 
(
	[OID] ASC
)WITH (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, SORT_IN_TEMPDB = OFF, IGNORE_DUP_KEY = OFF, ONLINE = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
ON [PRIMARY];

ALTER TABLE [nan].[Variable]  WITH CHECK ADD  CONSTRAINT [FK_Variable_LabelJob] FOREIGN KEY([LabelJobOID])
REFERENCES [nan].[LabelJob] ([OID]);

ALTER TABLE [nan].[Variable] CHECK CONSTRAINT [FK_Variable_LabelJob];

GO

-- Start archivation process
DECLARE @MinDate DATE = (
	SELECT TOP 1  MIN(CAST(rv.value AS DATETIME)) AS MinDate
	FROM sys.indexes i  
		INNER JOIN sys.partitions p ON i.object_id = p.object_id AND i.index_id=p.index_id  
		INNER JOIN sys.partition_schemes ps ON ps.data_space_id = i.data_space_id  
		INNER JOIN sys.partition_functions pf ON pf.function_id = ps.function_id 
		INNER JOIN sys.allocation_units au  ON au.container_id = p.hobt_id   
		INNER JOIN sys.filegroups fg ON fg.data_space_id = au.data_space_id  
		LEFT JOIN sys.partition_range_values rv ON rv.function_id = pf.function_id AND rv.boundary_id = p.partition_number 
	WHERE pf.name = 'ArchivatorPF'
		AND ps.Name = 'ArchivatorPS' 
		AND rv.value IS NOT NULL
	GROUP BY ps.Name, pf.name);

DECLARE @ArchivationInterval AS INT = (SELECT TOP 1 CAST(Value AS INT) FROM [nan].[Setting] WHERE Setting = 'ArchivationInterval');
DECLARE @ArchivationIntervalType AS BIT = (SELECT TOP 1 CAST(Value AS BIT) FROM [nan].[Setting] WHERE Setting = 'ArchivationIntervalType');
DECLARE @ArchivationIntervalMax AS INT = (SELECT TOP 1 CAST(Value AS INT) FROM [nan].[Setting] WHERE Setting = 'ArchivationIntervalMax');

DECLARE @Now AS DATETIME = GETUTCDATE();

DECLARE @ArchivationIntervalStart AS DATE = (
	SELECT 
		CASE @ArchivationIntervalType 
			WHEN 0 THEN DATEADD(DAY, -@ArchivationInterval * @ArchivationIntervalMax, @Now) 
			ELSE DATEADD(MONTH, -@ArchivationInterval * @ArchivationIntervalMax, DATEADD(DAY, -DAY(@Now) + 1, @Now)) 
		END);

IF @MinDate > @ArchivationIntervalStart
BEGIN
	RETURN;
END

IF NOT EXISTS (SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = N'ArchivePrintJob')
BEGIN
	SELECT TOP 1 *
	INTO [nan].[ArchivePrintJob]
	FROM [nan].[PrintJob];

	DELETE FROM [nan].[ArchivePrintJob];

	CREATE CLUSTERED INDEX [IXC_ArchivePrintJob_LastUsed] ON [nan].[ArchivePrintJob]
	(
		LastUsed
	) WITH (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, SORT_IN_TEMPDB = OFF, ONLINE = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
	ON ArchivatorPS(LastUsed);
END


IF NOT EXISTS (SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = N'ArchiveLabelJob')
BEGIN
	SELECT TOP 1 *
	INTO [nan].[ArchiveLabelJob]
	FROM [nan].[LabelJob];

	DELETE FROM [nan].[ArchiveLabelJob];
	
	CREATE CLUSTERED INDEX [IXC_ArchiveLabelJob_LastUsed] ON [nan].[ArchiveLabelJob]
	(
		LastUsed
	) WITH (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, SORT_IN_TEMPDB = OFF, ONLINE = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
	ON ArchivatorPS(LastUsed);
END
	
IF NOT EXISTS (SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = N'ArchiveVariable')
BEGIN
	SELECT TOP 1 *
	INTO [nan].[ArchiveVariable]
	FROM [nan].[Variable];

	DELETE FROM [nan].[ArchiveVariable];
	
	CREATE CLUSTERED INDEX [IXC_ArchiveVariable_LastUsed] ON [nan].[ArchiveVariable]
	(
		LastUsed
	) WITH (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, SORT_IN_TEMPDB = OFF, ONLINE = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
	ON ArchivatorPS(LastUsed);
END
	
IF NOT EXISTS (SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = N'ArchiveEvent')
BEGIN
	SELECT TOP 1 *
	INTO [nan].[ArchiveEvent]
	FROM [nan].[Event];

	DELETE FROM [nan].[ArchiveEvent];
	
	CREATE CLUSTERED INDEX [IXC_ArchiveEvent_LastUsed] ON [nan].[ArchiveEvent]
	(
		LastUsed
	) WITH (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, SORT_IN_TEMPDB = OFF, ONLINE = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
	ON ArchivatorPS(LastUsed);
END
	
IF NOT EXISTS (SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = N'ArchiveEventErrorDetails')
BEGIN
	SELECT TOP 1 *
	INTO [nan].[ArchiveEventErrorDetails]
	FROM [nan].[EventErrorDetails];

	DELETE FROM [nan].[ArchiveEventErrorDetails];
END
	
IF NOT EXISTS (SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = N'ArchiveEventErrorDetailsParameters')
BEGIN
	SELECT TOP 1 *
	INTO [nan].[ArchiveEventErrorDetailsParameters]
	FROM [nan].[EventErrorDetailsParameters];

	DELETE FROM [nan].[ArchiveEventErrorDetailsParameters];
END
	
IF NOT EXISTS (SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = N'ArchiveIssuedAlert')
BEGIN
	SELECT TOP 1 *
	INTO [nan].[ArchiveIssuedAlert]
	FROM [nan].[IssuedAlert];

	DELETE FROM [nan].[ArchiveIssuedAlert];
END

GO

EXEC [nan].[DropIndexes];

-- Event errors
DECLARE @EventErrorIds AS TABLE (Id INT);

SET IDENTITY_INSERT [nan].[ArchiveEventErrorDetails] ON;

INSERT INTO [nan].[ArchiveEventErrorDetails] (OID, ErrorId, EventOid)
OUTPUT inserted.OID INTO @EventErrorIds
SELECT ed.* 
FROM [nan].[EventErrorDetails] ed 
	INNER JOIN [nan].[Event] e ON ed.EventOid = e.OID 
WHERE $PARTITION.ArchivatorPF(e.LastUsed) = 1;

SET IDENTITY_INSERT [nan].[ArchiveEventErrorDetails] OFF;

-- Event error parameters
DECLARE @EventErrorParameterIds AS TABLE (Id INT);

SET IDENTITY_INSERT [nan].[ArchiveEventErrorDetailsParameters] ON;

INSERT INTO [nan].[ArchiveEventErrorDetailsParameters] (OID, EventErrorDetailsOId, Name, Value)
OUTPUT inserted.OID INTO @EventErrorParameterIds
SELECT edp.* 
FROM [nan].[EventErrorDetailsParameters] edp 
WHERE edp.EventErrorDetailsOId IN (SELECT Id FROM @EventErrorIds);

SET IDENTITY_INSERT [nan].[ArchiveEventErrorDetailsParameters] OFF;

-- Issued alerts
DECLARE @IssuedAlertIds AS TABLE (Id INT);

SET IDENTITY_INSERT [nan].[ArchiveIssuedAlert] ON;

INSERT INTO [nan].[ArchiveIssuedAlert] (OID, AlertOID, EventOID, Sent, Message)
OUTPUT inserted.OID INTO @IssuedAlertIds
SELECT ia.* 
FROM [nan].[IssuedAlert] ia 
	INNER JOIN [nan].[Event] e ON ia.EventOid = e.OID 
WHERE $PARTITION.ArchivatorPF(e.LastUsed) = 1;

SET IDENTITY_INSERT [nan].[ArchiveIssuedAlert] OFF;

DELETE FROM [nan].[EventErrorDetailsParameters] WHERE OID IN (SELECT Id FROM @EventErrorParameterIds);
DELETE FROM [nan].[EventErrorDetails] WHERE OID IN (SELECT Id FROM @EventErrorIds);
DELETE FROM [nan].[IssuedAlert] WHERE OID IN (SELECT Id FROM @IssuedAlertIds);

-- Switch the last partition. This one will contain archieved data
ALTER TABLE [nan].[PrintJob] SWITCH PARTITION 1 TO [nan].[ArchivePrintJob ]PARTITION 1;
ALTER TABLE [nan].[LabelJob] SWITCH PARTITION 1 TO [nan].[ArchiveLabelJob] PARTITION 1;
ALTER TABLE [nan].[Variable] SWITCH PARTITION 1 TO [nan].[ArchiveVariable] PARTITION 1;
ALTER TABLE [nan].[Event] SWITCH PARTITION 1 TO [nan].[ArchiveEvent] PARTITION 1;
	
EXEC [nan].[RecreateIndexes];
GO

ALTER PROCEDURE [nan].[ProcessArchivation]
AS   
BEGIN
	DECLARE @MinDate DATE = (SELECT TOP 1  MIN(CAST(rv.value AS DATETIME)) AS MinDate
		FROM sys.indexes i  
			INNER JOIN sys.partitions p ON i.object_id = p.object_id AND i.index_id = p.index_id  
			INNER JOIN sys.partition_schemes ps ON ps.data_space_id = i.data_space_id  
			INNER JOIN sys.partition_functions pf ON pf.function_id = ps.function_id
			INNER JOIN sys.allocation_units au  ON au.container_id = p.hobt_id   
			INNER JOIN sys.filegroups fg ON fg.data_space_id = au.data_space_id  
			LEFT JOIN sys.partition_range_values rv ON rv.function_id = pf.function_id AND rv.boundary_id = p.partition_number
		WHERE pf.name = 'ArchivatorPF' 
			AND ps.Name = 'ArchivatorPS' 
			AND rv.value IS NOT NULL
		GROUP BY ps.Name, pf.name);

	DECLARE @ArchivationInterval AS INT = (SELECT TOP 1 CAST(Value AS INT) FROM [nan].Setting WHERE Setting = 'ArchivationInterval');
	DECLARE @ArchivationIntervalType AS BIT = (SELECT TOP 1 CAST(Value AS BIT) FROM [nan].Setting WHERE Setting = 'ArchivationIntervalType');
	DECLARE @ArchivationIntervalMax AS INT = (SELECT TOP 1 CAST(Value AS INT) FROM [nan].Setting WHERE Setting = 'ArchivationIntervalMax');

	DECLARE @Now DATETIME = GETUTCDATE();

	DECLARE @ArchivationIntervalStart AS DATE = (
		SELECT 
			CASE @ArchivationIntervalType WHEN 0 THEN DATEADD(DAY, -@ArchivationInterval * @ArchivationIntervalMax, @Now) 
			ELSE DATEADD(MONTH, -@ArchivationInterval * @ArchivationIntervalMax, DATEADD(DAY, -DAY(@Now) + 1, @Now)) 
			END);

	IF @MinDate > @ArchivationIntervalStart
	BEGIN
		RETURN;
	END

	EXEC [nan].[DropIndexes];
	
	DECLARE @BeforeLastPartition INT = (
		SELECT COUNT(DISTINCT rv.Value)
		FROM sys.indexes i  
			INNER JOIN sys.partitions p ON i.object_id = p.object_id AND i.index_id = p.index_id  
			INNER JOIN sys.partition_schemes ps ON ps.data_space_id = i.data_space_id  
			INNER JOIN sys.partition_functions pf ON pf.function_id = ps.function_id  
			LEFT JOIN sys.partition_range_values rv ON rv.function_id = pf.function_id AND rv.boundary_id = p.partition_number
		WHERE pf.name = 'ArchivatorPF' 
			AND ps.Name = 'ArchivatorPS' 
			AND rv.value IS NOT NULL);
		
	-- Event errors
	DECLARE @EventErrorIds AS TABLE (Id INT);

	SET IDENTITY_INSERT [nan].[ArchiveEventErrorDetails] ON;

	INSERT INTO [nan].[ArchiveEventErrorDetails] (OID, ErrorId, EventOid)
	OUTPUT inserted.OID INTO @EventErrorIds
	SELECT ed.* 
	FROM [nan].[EventErrorDetails] ed 
		INNER JOIN [nan].[Event] e ON ed.EventOid = e.OID WHERE $PARTITION.ArchivatorPF(e.LastUsed) = 2;

	SET IDENTITY_INSERT [nan].[ArchiveEventErrorDetails] OFF;

	DECLARE @EventErrorParameterIds AS TABLE (Id INT);

	-- Event error details parameters
	SET IDENTITY_INSERT [nan].[ArchiveEventErrorDetailsParameters] ON;

	INSERT INTO [nan].[ArchiveEventErrorDetailsParameters] (OID, EventErrorDetailsOId, Name, Value)
	OUTPUT inserted.OID INTO @EventErrorParameterIds
	SELECT edp.* 
	FROM [nan].[EventErrorDetailsParameters] edp 
	WHERE edp.EventErrorDetailsOId IN (SELECT Id FROM @EventErrorIds);

	SET IDENTITY_INSERT [nan].[ArchiveEventErrorDetailsParameters] OFF;

	-- Issued alerts
	DECLARE @IssuedAlertIds AS TABLE (Id INT);

	SET IDENTITY_INSERT [nan].[ArchiveIssuedAlert] ON;

	INSERT INTO [nan].[ArchiveIssuedAlert] (OID, AlertOID, EventOID, Sent, Message)
	OUTPUT inserted.OID INTO @IssuedAlertIds
	SELECT ia.* 
	FROM [nan].[IssuedAlert] ia 
		INNER JOIN [nan].[Event] e ON ia.EventOid = e.OID 
	WHERE $PARTITION.ArchivatorPF(e.LastUsed) = 2;

	SET IDENTITY_INSERT [nan].[ArchiveIssuedAlert] OFF;
	
	DELETE FROM [nan].[EventErrorDetailsParameters] WHERE OID IN (SELECT Id FROM @EventErrorParameterIds);
	DELETE FROM [nan].[EventErrorDetails] WHERE OID IN (SELECT Id FROM @EventErrorIds);
	DELETE FROM [nan].[IssuedAlert] WHERE OID IN (SELECT Id FROM @IssuedAlertIds);

	-- Switch partition that is one before the last one. Last one contains archived data
	ALTER TABLE [nan].[PrintJob] SWITCH PARTITION 2 TO [nan].[ArchivePrintJob] PARTITION 2;
	ALTER TABLE [nan].[LabelJob] SWITCH PARTITION 2 TO [nan].[ArchiveLabelJob] PARTITION 2;
	ALTER TABLE [nan].[Variable] SWITCH PARTITION 2 TO [nan].[ArchiveVariable] PARTITION 2;
	ALTER TABLE [nan].[Event] SWITCH PARTITION 2 TO [nan].[ArchiveEvent] PARTITION 2;
	
	EXEC [nan].[RecreateIndexes];

	DECLARE @MaxDate DATETIME = (select TOP 1 MAX(CAST(rv.value AS DATETIME)) AS MaxDate
		FROM sys.indexes i  
			INNER JOIN sys.partitions p ON i.object_id = p.object_id AND i.index_id = p.index_id  
			INNER JOIN sys.partition_schemes ps ON ps.data_space_id = i.data_space_id  
			INNER JOIN sys.partition_functions pf ON pf.function_id = ps.function_id  
			INNER JOIN sys.allocation_units au  ON au.container_id = p.hobt_id   
			INNER JOIN sys.filegroups fg ON fg.data_space_id = au.data_space_id  
			LEFT JOIN sys.partition_range_values rv ON rv.function_id = pf.function_id AND rv.boundary_id = p.partition_number
		WHERE pf.name = 'ArchivatorPF' 
			AND ps.Name = 'ArchivatorPS' 
			AND rv.value IS NOT NULL
		GROUP BY ps.Name, pf.name);
	
	DECLARE @Query NVARCHAR(MAX) = 'ALTER PARTITION FUNCTION ArchivatorPF () MERGE RANGE (''' + CONVERT(NVARCHAR(50), @MinDate, 112) + ''');';
	EXEC sp_executesql @Query;

	ALTER PARTITION SCHEME ArchivatorPS NEXT USED [PRIMARY];

	IF @ArchivationIntervalType = 0
		SET @Query = 'ALTER PARTITION FUNCTION ArchivatorPF () SPLIT RANGE (''' + CONVERT(NVARCHAR(50), DATEADD(DAY, 1, @MaxDate), 112) + ''');';
	ELSE
		SET @Query = 'ALTER PARTITION FUNCTION ArchivatorPF () SPLIT RANGE (''' + CONVERT(NVARCHAR(50), DATEADD(MONTH, 1, @MaxDate), 112) + ''');';
	
	EXEC sp_executesql @Query;
END
GO